using namespace std;

#include <map>
#include <string>
#include <iostream>
#include <fstream>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "utils.h"
#include "genecode.h"

// ----------------------------------------------------------------------

static transl_table aa2nt[] = {
  {"A", "AAA"},  {"R", "AAT"},  {"N", "AAC"},  {"D", "AAG"},  {"C", "ATA"},
  {"Q", "ATT"},  {"E", "ATC"},  {"G", "ATG"},  {"H", "ACA"},  {"I", "ACT"},
  {"L", "ACC"},  {"K", "ACG"},  {"M", "AGA"},  {"F", "AGT"},  {"P", "AGC"},
  {"S", "AGG"},  {"T", "TAA"},  {"W", "TAT"},  {"Y", "TAC"},  {"V", "TAG"},
  {"*", "NNN"},  {NULL, NULL}
};


class codon{
public:
  char nt[4];

  codon(){ fixup(); };

  codon(const char* a){
    nt[0]=a[0]; nt[1]=a[1]; nt[2]=a[2]; fixup();
  }

  codon(const codon& a){
    nt[0]=a.nt[0]; nt[1]=a.nt[1]; nt[2]=a.nt[2];  fixup();
  }

  void operator=(codon a){
    nt[0]=a.nt[0]; nt[1]=a.nt[1]; nt[2]=a.nt[2];  fixup();
  }

  bool operator==(codon a){
    return a.nt[0]==nt[0] && a.nt[1]==nt[1] && a.nt[2]==nt[2];
  }

private:
  void fixup(){
    for(int i=0; i<3; i++)
      if(islower(nt[i])) nt[i]-=0x20;
    nt[3]=0;
  }

};

bool operator<(const codon& a, const codon& b){
  int aa, bb;
  aa = a.nt[0]*256*256 + a.nt[1]*256 + a.nt[2];
  bb = b.nt[0]*256*256 + b.nt[1]*256 + b.nt[2];
  return aa<bb;
}

// global stuff
map<codon, codon> nt2nt[transl_table_size];


// ----------------------------------------------------------------------

string reverse(const string& fwd){
  string rev="";

  for(int i=fwd.length()-1; i>=0; i--){
    char nt=fwd[i];
    switch(nt){
    case 'A': rev+='T'; break;
    case 'T': rev+='A'; break;
    case 'C': rev+='G'; break;
    case 'G': rev+='C'; break;

    case 'a': rev+='t'; break;
    case 't': rev+='a'; break;
    case 'c': rev+='g'; break;
    case 'g': rev+='c'; break;

    default:  rev+=nt;
    }
  }
  return rev;
}

void translate_nt(char* from, char* to, int table){
  map<codon,codon>::iterator ite=nt2nt[table].find(codon(from));
  
  if(ite != nt2nt[table].end())
    for(unsigned i=0; i<3; i++) to[i]=ite->second.nt[i];
   else 
    for(unsigned i=0; i<3; i++) to[i]='N';

  for(unsigned i=0; i<3; i++)
    if(islower(from[i])) to[i]+=0x20;
}


void translate(char*& seq, string& orig_seq, ofstream& ofs, int table){
  char ntbuf[4] = { 0,0,0,0 };
  
  seq = strdup(orig_seq.c_str());
  unsigned seqlen = strlen(seq);
  unsigned outlen = 3 * (orig_seq.length()/3) + 6;
  orig_seq="";

  // length in FASTA header
  ofs << ">" << outlen << endl;

  int codon_count = 0;
  for(int strand=0; strand<2; strand++){
    for(unsigned frame=0; frame<3; frame++){
      unsigned i;

      for(i=frame; i<=seqlen-3; i+=3){
	translate_nt(&seq[i], ntbuf, table);

	for(int ii=0; ii<3; ii++){
	  ofs << ntbuf[ii];
	  codon_count ++;
	  if(codon_count==60){
	    codon_count = 0;
	    ofs << endl;
	  }
	}
      }
      for(i=i-frame; i<outlen; i++){
	ofs << 'n';
	codon_count ++;
	if(codon_count==60){
	  codon_count = 0;
	  ofs << endl;
	}
      }
    }

    if(strand==0){
      string revtemp = reverse(seq);
      free(seq);
      seq = strdup(revtemp.c_str());
    } else {
      free(seq);
    }
  }

  ofs << endl;

}

int main(int argc, char *argv[]){
  // ----------------------------------------------------------------------
  // setup the translation table
  map<string, codon> aa2nt_map;

  transl_table *t;
  t=aa2nt;
  while(t->a != NULL){
    aa2nt_map[string(t->a)]=codon(t->b);
    t++;
  }

  for(int table=0; table<transl_table_size; table++){
    t=genetic_code[table];
    while(t->a != NULL){
      nt2nt[table][codon(t->a)]=aa2nt_map[string(t->b)];
      t++;
    }
  }
  
  // then test it.
//    string buf;
//    while(cin >> buf){
//      cout << nt2nt[codon(buf.c_str())].nt << endl;
//    }

  // ----------------------------------------------------------------------
  string infile = "/dev/stdin";
  int files=1;

  ofstream realseqs;
  string jobname = "test";
  string outpath = "output/";

  int ch;
  extern char *optarg;
  extern int  optind;

  static struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {"pattern", 1, 0, 'p'},
    {"name", 1, 0, 'n'},
    {"directory", 1, 0, 'd'},
    {0,0,0,0}
  };
  int longindex = 0;

  while ((ch=getopt_long(argc, argv, "hd:p:n:", long_options, &longindex)) != -1){
    switch(ch){
    case 'h':
      cerr << "usage: murasaki -MURASAKI_OPTIONS `tmurasakix -dDIRECTORY -nNAME -pPATTERN seq1 seq2...`\n";
      return -1;
      break;

    case 'd':
      cout << "--directory=" << optarg << endl;
      outpath = string(optarg);
      break;

    case 'p':
      cout << "-p";
      if(string(optarg).find_first_of(":") == string::npos){
	// pattern specified
	while(*optarg != 0){
	  cout << *optarg << *optarg << *optarg;
	  optarg++;
	}
      } else { // random pattern
	string patspec = string(optarg);
	size_t colon = patspec.find_first_of(":");
	string temppat = randpat(atoi(patspec.substr(0, colon).c_str()),
				 atoi(patspec.substr(colon+1).c_str()));
	for(string::iterator ite=temppat.begin(); ite!=temppat.end(); ite++)
	  cout << *ite << *ite << *ite;
      }
      cout << endl;
      break;

    case 'n':
      cout << "-n" << optarg << endl;
      jobname = string(optarg);
      break;
    }

  }

  mkdir(outpath.c_str(), 0755);
  realseqs.open(string(outpath + "/" + jobname + ".ntseqs").c_str());

  // tmurasakix options
  cout << "-R --hashskip=3" << endl;

  // prepare command line options
  argc -= optind;
  argv += optind;

  files = (argc==0) ? 1 : argc;

  for(int file=0; file<files; file++){
    int table = 11;
    if(argc==0 && file==0){ infile="-"; } else { infile=string(argv[file]); }
    if(infile=="-") infile="/dev/stdin";
    
    // if filename contains colon, it's NAME:TABLE style
    if(infile.find_first_of(":") != string::npos){
      size_t colonloc = infile.find_last_of(":");
      table = atoi(infile.substr(colonloc+1).c_str());
      infile = infile.substr(0, colonloc);
      //      cout << infile << " with transl table " << table << endl;
    }

    ifstream ifs;
    ifs.open(infile.c_str());
    if(!ifs){
      cerr << "file " << infile << " wasn't found\n";
      return -1;
    }
    realseqs << infile << endl;

    // output file
    string mainname;
    if(infile=="/dev/stdin"){
      mainname="stdin";
    } else {
      mainname=infile.substr(0, infile.find_last_of("."));
    }

    string outfile = "stdin.aa" + itos(table) + ".fa";
    if(infile!="/dev/stdin"){
      outfile = mainname + ".aa" + itos(table) + ".fa";
    }
    cout << outfile << endl;
  
    if(infile != "/dev/stdin" && check_timestamp(infile, outfile)){
      ifs.close();
      continue;
    }

    string lenfile = mainname;
    ofstream ofs;
    ofs.open(outfile.c_str());
  
    // time to go
    string buf;
    string temp_seq="";
    char *seq;
    bool reading = false;
    while(getline(ifs, buf)){
      if(buf.substr(0,6)=="ORIGIN" || buf[0]=='>'){
	if(reading) temp_seq += "NNNNNNNNNN";
	reading= true;
      } else if (reading) {
	for(unsigned i=0; i<buf.length(); i++){
	  if(isalpha(buf[i])){
	    char nt=buf[i];
	    if(! (nt=='A' || nt=='T' || nt=='C' || nt=='G' ||
		  nt=='a' || nt=='t' || nt=='c' || nt=='g')){
	      if(isupper(nt)) nt='N';
	      if(islower(nt)) nt='n';
	    }
	    temp_seq += nt;
	  }
	}
      }
    }
    translate(seq, temp_seq, ofs, table); // FIXME: the table #
    ofs.close();
 
    ifs.close();
  }

  return 0;
}
